package fpmon;

import fpmon.gui.FPGui;
import java.io.*;
import java.util.Properties;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.*;
import org.hyperic.sigar.Sigar;
import org.hyperic.sigar.SigarException;
import org.hyperic.sigar.SigarProxy;

/**
 * A simple app to display various system stats in a small window. This class 
 * merely sets up the GUI and scheduler, which do all the real work.
 *
 * @author Miles Asvachin
 */
public final class FPMon
{
	private static String dataDirectoryName = ".FPMon";
	private static String propertiesFileName = "FPMon.properties";
	private static String logFileName = "FPMon.log";
	
	/**
	 * @param args the command line arguments
	 */
	public static void main(String[] args) throws SigarException, IOException
	{
		final Properties properties = getProperties();
		final Logger logger = Logger.getLogger("");
		
		try
		{
			final Level logLevel = Level.parse(properties.getProperty("LogLevel"));
			logger.setLevel(logLevel);
		}
		catch(IllegalArgumentException ex)
		{
			logger.setLevel(Level.WARNING);
			logger.log(Level.WARNING, "Defaulting to log level INFO", ex);
		}
		
		final File logFile = new File(getDataDirectory(), logFileName);
		final FileHandler logFileHandler = new FileHandler(logFile.getPath());
		final Formatter logFormatter = new SimpleFormatter();
		logFileHandler.setFormatter(logFormatter);
		logger.addHandler(logFileHandler);
		
		logger.log(Level.INFO, "Started FPMon");
		
		//Make sure the program quits when there is an exception, rather than 
		//the default behaviour of only the affected thread being terminated.
		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler()
		{
			@Override
			public void uncaughtException(final Thread t, final Throwable ex)
			{
				logger.log(Level.SEVERE, "Uncaught exception", ex);
				System.exit(1);
			}
		});
		
		javax.swing.SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				final int initialDelay = Integer.parseInt(properties.getProperty("InitialDelay"));
				final int updatePeriod = Integer.parseInt(properties.getProperty("UpdatePeriod"));
				final SigarProxy sigar = new Sigar();
				final FPGui gui = new FPGui(sigar, properties);
				final ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
				scheduler.scheduleWithFixedDelay(gui.getUpdater(), initialDelay, updatePeriod, TimeUnit.MILLISECONDS);
			}
		});
	}
	
	/**
	 * @return the FPMon data directory where settings and stuff can be written.
	 */
	private static File getDataDirectory() throws IOException
	{
		final String userHome = System.getProperty("user.home", ".");
		final File homeDirectory = new File(userHome);
		final File dataDirectory = new File(homeDirectory, dataDirectoryName);
		
		if(!dataDirectory.exists())
		{
			if(!dataDirectory.mkdir())
			{
				throw new IOException();
			}
		}
		
		return dataDirectory;
	}
	
	/**
	 * Loads the properties for FPMon. If a user properties file is present, those
	 * properties will override the default properties.
	 * 
	 * @return the properties
	 * @throws IOException if the default properties could not be loaded
	 */
	private static Properties getProperties() throws IOException
	{
		final File dataDirectory = getDataDirectory();
		final File propertiesFile = new File(dataDirectory, propertiesFileName);

		final Properties properties;
		
		if(!propertiesFile.exists())
		{
			Logger.getLogger(FPMon.class.getName()).log(Level.INFO, "Settings file not found; creating one with defaults.");
			
			copyDefaultPropertiesTo(propertiesFile);
		}
		
		if(propertiesFile.exists())
		{
			final InputStream propertiesIs = ClassLoader.getSystemResourceAsStream("fpmon/default.properties");
			final Properties defaultProperties = new Properties();
			defaultProperties.load(propertiesIs);
			properties = new Properties(defaultProperties);

			try (final InputStream is = new FileInputStream(propertiesFile))
			{
				properties.load(is);
			}
			catch(IOException ex)
			{
				Logger.getLogger(FPMon.class.getName()).log(Level.SEVERE, "Couldn't load settings file; using defaults.", ex);
			}
		}
		else
		{
			throw new IOException("No properties file");
		}
		
		return properties;
	}
	
	private static void copyDefaultPropertiesTo(final File destinationFile) throws FileNotFoundException, IOException
	{
		if(!destinationFile.exists())
		{
			if(!destinationFile.createNewFile())
			{
				throw new IOException("Failed to create properties file: " + destinationFile);
			}
		}
		
		final InputStream sourceIS = ClassLoader.getSystemResourceAsStream("fpmon/default.properties");
		
		try (final OutputStream destinationOS = new FileOutputStream(destinationFile))
		{
			final byte[] buffer = new byte[1024];
			int length;
			
			while ((length = sourceIS.read(buffer)) > 0)
			{
				destinationOS.write(buffer, 0, length);
			}
			
			sourceIS.close();
		}
	}
}
