package fpmon.gui;

import fpmon.stats.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import org.hyperic.sigar.FileSystem;
import org.hyperic.sigar.NetInterfaceConfig;
import org.hyperic.sigar.SigarException;
import org.hyperic.sigar.SigarProxy;

/**
 * Creates various chart displays and assembles them into a single frame.
 *
 * @author Miles Asvachin
 */
public final class FPGui
{
	private final static Dimension defaultSize = new Dimension(640, 480);
	private final StatUpdater updater;
	
	/**
	 * @param newSigar the sigar used to collect stats
	 * @param updatePeriod time in milliseconds between updates. This is only 
	 * used for initial calibration of axes; GUI updates need to be triggered 
	 * externally.
	 */
	public FPGui(final SigarProxy sigar, final Properties properties)
	{		
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}
		catch(ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException ex)
		{
			Logger.getLogger(FPGui.class.getName()).log(Level.WARNING, null, ex);
		}

		final JComponent contentPane = new JPanel();
		contentPane.setLayout(new GridLayout(3, 2));
		contentPane.setBackground(Color.BLACK);
		contentPane.setOpaque(true);
		
		final Collection<StatCollector> collectors = new LinkedList<>();
		final int updatePeriod = Integer.parseInt(properties.getProperty("UpdatePeriod"));
		
		try
		{
			collectors.add(new GeneralStatCollector(createStatPanel(contentPane), sigar, properties));
		}
		catch(MalformedURLException ex)
		{
			Logger.getLogger(FPGui.class.getName()).log(Level.SEVERE, "Couldn't create general stat collector", ex);
		}
		
		collectors.add(new CpuStatCollector(createStatPanel(contentPane), updatePeriod, sigar));
		collectors.add(new MemStatCollector(createStatPanel(contentPane), updatePeriod, sigar));

		try
		{
			final String netInterface = getNetInterface(sigar);
			collectors.add(new NetStatCollector(createStatPanel(contentPane), updatePeriod, netInterface, sigar));
		}
		catch(UnknownHostException | SigarException ex)
		{
			Logger.getLogger(FPGui.class.getName()).log(Level.SEVERE, "Couldn't create net stat collector", ex);
		}
		
		try
		{
			final FileSystem[] fileSystems = sigar.getFileSystemList();
			collectors.add(new DiscStatCollector(createStatPanel(contentPane), updatePeriod, fileSystems[0].getDevName(), sigar));
			
			if(fileSystems.length > 1)
			{
				collectors.add(new DiscStatCollector(createStatPanel(contentPane), updatePeriod, fileSystems[1].getDevName(), sigar));
			}
		}
		catch(SigarException ex)
		{
			Logger.getLogger(FPGui.class.getName()).log(Level.SEVERE, "Couldn't create disc stat collector", ex);
		}

		updater = new StatUpdater(sigar, collectors);

		JFrame.setDefaultLookAndFeelDecorated(true);
		final int screenDevice = Integer.parseInt(properties.getProperty("ScreenDevice"));
		final boolean fullscreen = Boolean.parseBoolean(properties.getProperty("Fullscreen"));
		final boolean alwaysOnTop = Boolean.parseBoolean(properties.getProperty("AlwaysOnTop"));
		final GraphicsConfiguration initialGraphicsConfiguration = getGraphicsConfiguration(screenDevice);
		final Rectangle defaultBounds = getDefaultBounds(initialGraphicsConfiguration);
		final Rectangle initialBounds = fullscreen ? getFullscreenBounds(initialGraphicsConfiguration) : defaultBounds;
		final JFrame frame = createFrame(contentPane, !fullscreen, initialBounds, alwaysOnTop);
		
		final FPMonMouseListener mouseListener = new FPMonMouseListener(frame, defaultBounds);
		frame.addMouseListener(mouseListener);
		frame.addMouseMotionListener(mouseListener);
	}
	
	/**
	 * @param screenDevice the device to get the config for
	 * @return the config of the given device, or the config of the default device
	 * if the given device is invalid
	 */
	private GraphicsConfiguration getGraphicsConfiguration(final int screenDevice)
	{
		final GraphicsEnvironment localGraphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
		final GraphicsDevice[] graphicsDevices = localGraphicsEnvironment.getScreenDevices();
		final GraphicsDevice graphicsDevice;
		
		if(screenDevice >= 0 && graphicsDevices.length > screenDevice)
		{
			graphicsDevice = graphicsDevices[screenDevice];
		}
		else
		{
			graphicsDevice = localGraphicsEnvironment.getDefaultScreenDevice();
			Logger.getLogger(FPGui.class.getName()).log(Level.WARNING, "Invalid ScreenDevice specified: {0}", screenDevice);
		}
		
		return graphicsDevice.getDefaultConfiguration();
	}
	
	/**
	 * @param graphicsConfiguration the config to place the bounds in
	 * @return bounds of defaultSize in the middle of the screen
	 */
	private Rectangle getDefaultBounds(final GraphicsConfiguration graphicsConfiguration)
	{
		final Rectangle screenBounds = graphicsConfiguration.getBounds();
		final int locationX = ((screenBounds.width - defaultSize.width) / 2) + screenBounds.x;
		final int locationY = ((screenBounds.height - defaultSize.height) / 2) + screenBounds.y;
		final Point location = new Point(locationX, locationY);
		
		return new Rectangle(location, defaultSize);
	}
	
	/**
	 * @param graphicsConfiguration the config to place the bounds in
	 * @return bounds that precisely fill the screen
	 */
	private Rectangle getFullscreenBounds(final GraphicsConfiguration graphicsConfiguration)
	{
		final Rectangle screenBounds = graphicsConfiguration.getBounds();
		final Dimension size = screenBounds.getSize();
		final Point location = screenBounds.getLocation();
		
		return new Rectangle(location, size);
	}
	
	/**
	 * Creates a frame, sets it up and displays it.
	 * 
	 * @param contentPane the pane the frame will contain
	 * @param decorated whether the frame should be decorated or not. Should 
	 * usually be false if the frame is fullscreen, true otherwise.
	 * @param bounds the size and location of the frame
	 * @return the created frame
	 */
	private JFrame createFrame(final Container contentPane, final boolean decorated, final Rectangle bounds, final boolean alwaysOnTop)
	{
		JFrame.setDefaultLookAndFeelDecorated(true);
		final JFrame frame = new JFrame("Front Panel Monitor");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setUndecorated(!decorated);
		frame.setIconImage(new ImageIcon(ClassLoader.getSystemResource("fpmon/gui/Icon32.png"), "Icon32").getImage());
		frame.setContentPane(contentPane);
		frame.setAlwaysOnTop(alwaysOnTop);

		if(bounds != null)
		{
			frame.setBounds(bounds);
		}
		else
		{
			frame.pack();
		}
		
		frame.setVisible(true);
		
		return frame;
	}
	
	/**
	 * Tries to find the main network interface.
	 */
	private String getNetInterface(final SigarProxy sigar) throws UnknownHostException, SigarException
	{
		String result = null;
		final InetAddress localHost = InetAddress.getLocalHost();
		final String address = localHost.getHostAddress();
		final String[] netInterfaces = sigar.getNetInterfaceList();

		for(String netInterface : netInterfaces)
		{
			final NetInterfaceConfig config = sigar.getNetInterfaceConfig(netInterface);

			if(config.getAddress().equals(address))
			{
				result = config.getName();
				break;
			}
		}
		
		return result;
	}
	
	/**
	 * Creates a panel suitable for a StatCollector to display in, and adds it 
	 * to the content pane.
	 */
	private JComponent createStatPanel(final JComponent contentPane)
	{
		final JComponent statPanel = new JPanel();
		contentPane.add(statPanel);
		
		return statPanel;
	}
	
	public StatUpdater getUpdater()
	{
		return updater;
	}

	/**
	 * Handles toggling between fullscreen and windowed modes, and hiding the cursor when the mouse stops moving.
	 */
	private static class FPMonMouseListener implements MouseListener, MouseMotionListener, ActionListener
	{
		private static int mouseHideWaitTime = 1000;
		private final Cursor blankCursor;
		private final Timer moveTimer;
		
		private JFrame frame;
		/**
		 * Where the window was before it was maximized, so it can be restored 
		 * to the same place later.
		 */
		private Rectangle oldBounds;

		public FPMonMouseListener(final JFrame newFrame, final Rectangle defaultBounds)
		{
			final BufferedImage cursorImg = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB);
			blankCursor = Toolkit.getDefaultToolkit().createCustomCursor(cursorImg, new Point(0, 0), "blank cursor");

			frame = newFrame;
			oldBounds = defaultBounds;
			
			moveTimer = new Timer(mouseHideWaitTime, this);
			moveTimer.setRepeats(false);
			moveTimer.start();
		}

		@Override
		public void mouseClicked(final MouseEvent e)
		{
			if(e.getClickCount() == 2)
			{
				if(frame.isUndecorated())
				{
					//Restore - put the frame decorations back and restore the
					//bounds of the window from before it was maximized.
					frame.setVisible(false);
					frame.dispose();
					frame.setUndecorated(false);
					frame.setExtendedState(Frame.NORMAL);
					frame.setBounds(oldBounds);
				}
				else
				{
					//Maximize - remove the frame decorations and exactly fill the current monitor.
					oldBounds = frame.getBounds();
					frame.setVisible(false);
					frame.dispose();
					frame.setUndecorated(true);
					frame.setExtendedState(Frame.MAXIMIZED_BOTH);
				}
				
				frame.setVisible(true);
			}
		}
		
		private void checkMoveTimer()
		{
			if(moveTimer.isRunning())
			{
				moveTimer.restart();
			}
			else
			{
				frame.getContentPane().setCursor(null);
				moveTimer.start();
			}
		}

		@Override
		public void mousePressed(MouseEvent e)
		{
			checkMoveTimer();
		}

		@Override
		public void mouseReleased(MouseEvent e){}

		@Override
		public void mouseEntered(MouseEvent e)
		{
			checkMoveTimer();
		}

		@Override
		public void mouseExited(MouseEvent e)
		{
			if(moveTimer.isRunning())
			{
				moveTimer.stop();
			}
		}

		@Override
		public void mouseDragged(MouseEvent e)
		{
			checkMoveTimer();
		}

		@Override
		public void mouseMoved(MouseEvent e)
		{
			checkMoveTimer();
		}

		@Override
		public void actionPerformed(ActionEvent e)
		{
			frame.getContentPane().setCursor(blankCursor);
		}
	}
}
